from pydantic import model_validator

from stac_pydantic.item import Item as BaseItem
from stac_pydantic.links import Relations


class Item(BaseItem):
    @model_validator(mode="after")
    def required_links(self) -> "Item":
        links_rel = []
        for link in self.links.root:
            links_rel.append(link.rel)

        required_rels = [Relations.root, Relations.self, Relations.collection]

        for rel in required_rels:
            assert (
                rel in links_rel
            ), f"STAC API FEATURE conform Item pages must include a `{rel}` link."

        return self



from typing import List, Optional

from pydantic import model_validator

from stac_pydantic.api.collection import Collection
from stac_pydantic.api.links import Links
from stac_pydantic.links import Relations
from stac_pydantic.shared import StacBaseModel


class Collections(StacBaseModel):
    """
    https://github.com/radiantearth/stac-api-spec/tree/v1.0.0/ogcapi-features#endpoints
    https://github.com/radiantearth/stac-api-spec/tree/v1.0.0/ogcapi-features#collections-collections
    """

    links: Links
    collections: List[Collection]
    numberMatched: Optional[int] = None
    numberReturned: Optional[int] = None

    @model_validator(mode="after")
    def required_links(self) -> "Collections":
        links_rel = []
        for link in self.links.root:
            links_rel.append(link.rel)

        required_rels = [Relations.root, Relations.self]

        for rel in required_rels:
            assert (
                rel in links_rel
            ), f"STAC API COLLECTIONS conform Collections pages must include a `{rel}` link."

        return self



from typing import List

from pydantic import AnyUrl, HttpUrl, model_validator

from stac_pydantic.api.links import Links
from stac_pydantic.api.version import STAC_API_VERSION
from stac_pydantic.catalog import Catalog
from stac_pydantic.links import Relations


class LandingPage(Catalog):
    """
    https://github.com/radiantearth/stac-api-spec/tree/v1.0.0/core
    https://github.com/radiantearth/stac-api-spec/tree/v1.0.0/ogcapi-features#landing-page-
    https://github.com/radiantearth/stac-api-spec/tree/v1.0.0/item-search#link-relations
    """

    conformsTo: List[AnyUrl] = [
        HttpUrl("https://api.stacspec.org/v1.0.0/core"),
        HttpUrl("http://www.opengis.net/spec/ogcapi-features-1/1.0/conf/core"),
    ]
    links: Links

    @model_validator(mode="after")
    def required_links(self) -> "LandingPage":
        links_rel = []
        for link in self.links.root:
            links_rel.append(link.rel)

        required_core_rels = [Relations.root, Relations.self, Relations.service_desc]

        for rel in required_core_rels:
            assert (
                rel in links_rel
            ), f"STAC API conform Landing pages must include a `{rel}` link."

        if (
            AnyUrl(f"https://api.stacspec.org/v{STAC_API_VERSION}/collections")
            in self.conformsTo
        ):
            required_collections_rels = [Relations.data]
            for rel in required_collections_rels:
                assert (
                    rel in links_rel
                ), f"STAC API COLLECTION conform Landing pages must include a `{rel}` link."

        if (
            AnyUrl(f"https://api.stacspec.org/v{STAC_API_VERSION}/item-search")
            in self.conformsTo
        ):
            required_feature_rels = [Relations.search]
            for rel in required_feature_rels:
                assert (
                    rel in links_rel
                ), f"STAC API ITEM SEARCH conform Landing pages must include a `{rel}` link."

        return self



from enum import auto
from typing import Iterator, List, Optional
from urllib.parse import urljoin

from pydantic import ConfigDict, Field, RootModel

from stac_pydantic.shared import MimeTypes, StacBaseModel
from stac_pydantic.utils import AutoValueEnum


class Link(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#link-object
    """

    href: str = Field(..., alias="href", min_length=1)
    rel: str = Field(..., alias="rel", min_length=1)
    type: Optional[MimeTypes] = None
    title: Optional[str] = None

    # Label extension
    label: Optional[str] = Field(default=None, alias="label:assets")
    model_config = ConfigDict(use_enum_values=True, extra="allow")

    def resolve(self, base_url: str) -> None:
        """resolve a link to the given base URL"""
        self.href = urljoin(base_url, self.href)


class Links(RootModel[List[Link]]):
    root: List[Link]

    def link_iterator(self) -> Iterator[Link]:
        """Produce iterator to iterate through links"""
        return iter(self.root)

    def resolve(self, base_url: str) -> None:
        """resolve all links to the given base URL"""
        for link in self.link_iterator():
            link.resolve(base_url)

    def append(self, link: Link) -> None:
        self.root.append(link)

    def __len__(self) -> int:
        return len(self.root)

    def __getitem__(self, idx: int) -> Link:
        return self.root[idx]


class Relations(str, AutoValueEnum):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#relation-types
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/best-practices.md#using-relation-types
    """

    self = auto()
    root = auto()
    parent = auto()
    child = auto()
    children = auto()
    items = auto()
    license = auto()
    derived_from = auto()
    collection = auto()
    alternate = auto()
    previous = auto()
    prev = auto()
    next = auto()
    conformance = auto()
    docs = auto()
    tiles = auto()
    search = auto()
    preview = auto()
    canonical = auto()
    via = auto()
    data = auto()
    service_desc = "service-desc"
    service_doc = "service-doc"
    queryables = "http://www.opengis.net/def/rel/ogc/1.0/queryables"



from typing import Optional, Sequence

from geojson_pydantic import FeatureCollection

from stac_pydantic.item import Item
from stac_pydantic.links import Links
from stac_pydantic.shared import StacBaseModel


class ItemCollection(FeatureCollection, StacBaseModel):
    """
    This is a less strict implementation of ItemCollection.
    It only implements the FeatureCollection Specs
    not enforcing required links as specified in STAC-API-FEATURES specs.
    Use `stac_pydantic.api.ItemCollection` to enforce link relationships and extra fields.
    """

    features: Sequence[Item]  # type: ignore
    links: Optional[Links] = None



"""STAC API models."""

import warnings

from .collection import Collection
from .collections import Collections
from .conformance import Conformance
from .item import Item
from .item_collection import ItemCollection
from .landing import LandingPage
from .search import Search

__all__ = [
    "Collection",
    "Collections",
    "Conformance",
    "Item",
    "ItemCollection",
    "LandingPage",
    "Search",
]


# TODO: remove in 4.0
def __getattr__(name):
    if name == "ConformanceClasses":
        warnings.warn(
            "Class `ConformanceClasses` has been renamed to `Conformance`. Please use the new name. The old alias will be removed in a future version.",
            DeprecationWarning,
            stacklevel=2,
        )
        return Conformance

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Union

from pydantic import AfterValidator, Field, conlist
from typing_extensions import Annotated

from stac_pydantic.catalog import _Catalog
from stac_pydantic.shared import (
    Asset,
    BBox,
    NumType,
    Provider,
    StacBaseModel,
    UtcDatetime,
    validate_bbox,
)

if TYPE_CHECKING:
    StartEndTime = List[Union[UtcDatetime, None]]
    TInterval = List[StartEndTime]
else:
    StartEndTime = conlist(Union[UtcDatetime, None], min_length=2, max_length=2)
    TInterval = conlist(StartEndTime, min_length=1)


def validate_bbox_interval(v: List[BBox]) -> List[BBox]:  # noqa: C901
    ivalues = iter(v)

    # The first time interval always describes the overall spatial extent of the data.
    overall_bbox = next(ivalues, None)
    if not overall_bbox:
        return v

    assert validate_bbox(overall_bbox)

    if len(overall_bbox) == 4:
        xmin, ymin, xmax, ymax = overall_bbox
    else:
        xmin, ymin, _, xmax, ymax, _ = overall_bbox

    crossing_antimeridian = xmin > xmax
    for bbox in ivalues:
        error_msg = ValueError(
            f"`BBOX` {bbox} not fully contained in `Overall BBOX` {overall_bbox}"
        )
        _ = validate_bbox(bbox)

        if len(bbox) == 4:
            xmin_sub, ymin_sub, xmax_sub, ymax_sub = bbox
        else:
            xmin_sub, ymin_sub, _, xmax_sub, ymax_sub, _ = bbox

        if not ((ymin_sub >= ymin) and (ymax_sub <= ymax)):
            raise error_msg

        sub_crossing_antimeridian = xmin_sub > xmax_sub
        if not crossing_antimeridian and sub_crossing_antimeridian:
            raise error_msg

        elif crossing_antimeridian:
            #                           Antimeridian
            #     0                     + 180 │ - 180                            0
            #     │  [176,1,179,3]            │                                  │
            #     │       │                   │                                  │
            #     │       │                   │                                  │
            #     │       │                   │                                  │     [-178,1,-176,3]
            #     │       │  ┌─────────────────────────────────────────┐         │           │
            #     │       │  │       xmax_sub │               xmax_sub │         │           │
            #     │       │  │  ┌──────|      │        ┌─────────|     │         │           │
            #     │       └──│──►  2   │      │        │    3    │     │         │           │
            #     |          │  │      │      │        │         │◄────│─────────┼───────────┘
            #     │          │  |──────┘      │        |─────────┘     │         │
            #     │          │xmin_sub        │    xmin_sub            │         │         0
            #   ──┼──────────│────────────────┼────────────────────────│─────────┼──────────
            #     │          │                │    xmax_sub(-179)      │         │
            #     │          │         ┌──────────────|                │         │
            #     │          │         │      │       │                │         │
            #     │          │         │      │  1    │                │         │
            #     |          │         │      │       │◄────────┐      │◄────────┼─────── [175,-3,-174,5]
            #     │          │         │      │       │         │      │         │
            #     │          │         |──────────────┘         │      │         │
            #     │          │   xmin_sub(179)│                 │      │         │
            #     │          |──────────────────────────────────┼──────|         │
            #     │      xmin(174)            │                 │ xmax(-174)     │
            #     │                           │                 │                │
            #     │                           │                 │                │
            #     │                           │                 │                │
            #     │                           │          [179,-2,-179,-1]        │

            # Case 1
            if sub_crossing_antimeridian:
                if not (xmin_sub > xmin and xmax_sub < xmax):
                    raise error_msg

            # Case 2: if sub-sequent has lon > 0 (0 -> 180 side), then we must check if
            # its min lon is < to the western lon (xmin for bbox crossing antimeridian limit)
            # of the overall bbox (on 0 -> +180 side)
            elif xmin_sub >= 0 and xmin_sub < xmin:
                raise error_msg

            # Case 3: if sub-sequent has lon < 0 (-180 -> 0 side), then we must check if
            # its max lon is > to the eastern lon (xmax for bbox crossing antimeridian limit)
            #  of the overall bbox (on -180 -> 0 side)
            elif xmin_sub <= 0 and xmax_sub > xmax:
                raise error_msg

        else:
            if not ((xmin_sub >= xmin) and (xmax_sub <= xmax)):
                raise error_msg

    return v


def validate_time_interval(v: TInterval) -> TInterval:  # noqa: C901
    ivalues = iter(v)

    # The first time interval always describes the overall temporal extent of the data.
    overall_interval = next(ivalues, None)
    if not overall_interval:
        return v

    start, end = overall_interval
    if start and end:
        if start > end:
            raise ValueError(f"`Start` time {start} older than `End` time {end}")

    # All subsequent time intervals can be used to provide a more precise
    # description of the extent and identify clusters of data.
    for s, e in ivalues:
        if s and e:
            if s > e:
                raise ValueError(f"`Start` time {s} older than `End` time {e}")

        if start and s:
            if start > s:
                raise ValueError(
                    f"`Overall Start` time {start} older than `Start` time {s}"
                )

        if end and e:
            if e > end:
                raise ValueError(
                    f"`End` time {e} older than `Overall Start` time {end}"
                )

    return v


class SpatialExtent(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#spatial-extent-object
    """

    bbox: Annotated[List[BBox], AfterValidator(validate_bbox_interval)]


class TimeInterval(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#temporal-extent-object
    """

    interval: Annotated[TInterval, AfterValidator(validate_time_interval)]


class Extent(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#extent-object
    """

    spatial: SpatialExtent
    temporal: TimeInterval


class Range(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#stats-object
    """

    minimum: Union[NumType, str]
    maximum: Union[NumType, str]


class Collection(_Catalog):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md
    """

    assets: Optional[Dict[str, Asset]] = None
    license: str = Field(..., alias="license", min_length=1)
    extent: Extent
    title: Optional[str] = None
    keywords: Optional[List[str]] = None
    providers: Optional[List[Provider]] = None
    summaries: Optional[Dict[str, Union[Range, List[Any], Dict[str, Any]]]] = None
    type: Literal["Collection"]



from datetime import datetime as dt
from typing import Any, Dict, List, Optional, Union

from geojson_pydantic.geometries import (
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)
from pydantic import AfterValidator, BaseModel, Field, model_validator
from typing_extensions import Annotated

from stac_pydantic.api.extensions.fields import FieldsExtension
from stac_pydantic.api.extensions.query import Operator
from stac_pydantic.api.extensions.sort import SortExtension

# TODO: remove in 4.0
from stac_pydantic.shared import SearchDatetime  # noqa
from stac_pydantic.shared import (
    BBox,
    str_to_datetimes,
    validate_bbox,
    validate_datetime,
)

Intersection = Union[
    Point,
    MultiPoint,
    LineString,
    MultiLineString,
    Polygon,
    MultiPolygon,
    GeometryCollection,
]


class Search(BaseModel):
    """
    The base class for STAC API searches.

    https://github.com/radiantearth/stac-api-spec/blob/v1.0.0/item-search/README.md#query-parameter-table
    """

    collections: Optional[List[str]] = None
    ids: Optional[List[str]] = None
    bbox: Annotated[Optional[BBox], AfterValidator(validate_bbox)] = None
    intersects: Optional[Intersection] = None
    datetime: Annotated[Optional[str], AfterValidator(validate_datetime)] = None
    limit: Optional[int] = 10

    @property
    def start_date(self) -> Optional[dt]:
        start_date: Optional[dt] = None
        if self.datetime:
            start_date = str_to_datetimes(self.datetime)[0]
        return start_date

    @property
    def end_date(self) -> Optional[dt]:
        end_date: Optional[dt] = None
        if self.datetime:
            dates = str_to_datetimes(self.datetime)
            end_date = dates[0] if len(dates) == 1 else dates[1]
        return end_date

    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-validators for more information.
    @model_validator(mode="before")
    def validate_spatial(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get("intersects") and values.get("bbox") is not None:
            raise ValueError("intersects and bbox parameters are mutually exclusive")
        return values

    @property
    def spatial_filter(self) -> Optional[Intersection]:
        """Return a geojson-pydantic object representing the spatial filter for the search request.

        Check for both because the ``bbox`` and ``intersects`` parameters are mutually exclusive.
        """
        if self.bbox:
            return Polygon.from_bounds(*self.bbox)
        if self.intersects:
            return self.intersects
        else:
            return None


class ExtendedSearch(Search):
    """
    STAC API search with extensions enabled.
    """

    field: Optional[FieldsExtension] = Field(None, alias="fields")
    query: Optional[Dict[str, Dict[Operator, Any]]] = None
    sortby: Optional[List[SortExtension]] = None



import json

from stac_pydantic import ItemCollection
from stac_pydantic.version import STAC_VERSION

from .conftest import dict_match, request

ITEM_COLLECTION = "itemcollection-sample-full.json"
PATH = ["tests", "example_stac"]


def test_item_collection():
    test_item_coll = request(ITEM_COLLECTION, PATH)

    for feat in test_item_coll["features"]:
        feat["stac_version"] = STAC_VERSION

    valid_item_coll = ItemCollection(**test_item_coll).model_dump()
    for idx, feat in enumerate(test_item_coll["features"]):
        dict_match(feat, valid_item_coll["features"][idx])


def test_to_json():
    test_item = request(ITEM_COLLECTION, PATH)
    for feat in test_item["features"]:
        feat["stac_version"] = STAC_VERSION

    validated = ItemCollection(**test_item)
    dict_match(json.loads(validated.model_dump_json()), validated.model_dump())



from datetime import datetime as dt
from datetime import timezone
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Tuple, Union, cast
from warnings import warn

from pydantic import (
    AfterValidator,
    AwareDatetime,
    BaseModel,
    ConfigDict,
    Field,
    TypeAdapter,
    model_validator,
)
from typing_extensions import Annotated, Self

from stac_pydantic.utils import AutoValueEnum

NumType = Union[float, int]
BBox = Union[
    Tuple[NumType, NumType, NumType, NumType],  # 2D bbox
    Tuple[NumType, NumType, NumType, NumType, NumType, NumType],  # 3D bbox
]

SEMVER_REGEX = r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"

# Allows for some additional flexibility in the input datetime format. As long as
# the input value has timezone information, it will be converted to UTC timezone.
UtcDatetime = Annotated[
    # Input value must be in a format which has timezone information
    AwareDatetime,
    # Convert the input value to UTC timezone
    AfterValidator(lambda d: d.astimezone(timezone.utc)),
]

SearchDatetime = TypeAdapter(Optional[UtcDatetime])


class MimeTypes(str, Enum):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/item-spec.md#media-types
    """

    # Raster
    geotiff = "image/tiff; application=geotiff"
    cog = "image/tiff; application=geotiff; profile=cloud-optimized"
    jp2 = "image/jp2"
    png = "image/png"
    jpeg = "image/jpeg"
    # Vector
    geojson = "application/geo+json"
    geojsonseq = "application/geo+json-seq"
    geopackage = "application/geopackage+sqlite3"
    kml = "application/vnd.google-earth.kml+xml"
    kmz = "application/vnd.google-earth.kmz"
    pbf = "application/x-protobuf"
    mvt = "application/vnd.mapbox-vector-tile"
    # Others
    hdf = "application/x-hdf"
    hdf5 = "application/x-hdf5"
    xml = "application/xml"
    json = "application/json"
    ndjson = "application/ndjson"
    html = "text/html"
    text = "text/plain"
    openapi = "application/vnd.oai.openapi+json;version=3.0"
    openapi_yaml = "application/vnd.oai.openapi;version=3.0"
    jsonschema = "application/schema+json"
    pdf = "application/pdf"
    csv = "text/csv"
    parquet = "application/vnd.apache.parquet"
    octet_stream = "application/octet-stream"


class AssetRoles(str, AutoValueEnum):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/extensions/asset/README.md
    """

    thumbnail = auto()
    overview = auto()
    data = auto()
    metadata = auto()


class ProviderRoles(str, AutoValueEnum):
    licensor = auto()
    producer = auto()
    processor = auto()
    host = auto()


class StacBaseModel(BaseModel):
    def to_dict(
        self, by_alias: bool = True, exclude_unset: bool = True, **kwargs: Any
    ) -> Dict[str, Any]:
        warn(
            "`to_dict` method is deprecated. Use `model_dump` instead",
            DeprecationWarning,
        )
        return self.model_dump(by_alias=by_alias, exclude_unset=exclude_unset, **kwargs)

    def to_json(
        self, by_alias: bool = True, exclude_unset: bool = True, **kwargs: Any
    ) -> str:
        warn(
            "`to_json` method is deprecated. Use `model_dump_json` instead",
            DeprecationWarning,
        )
        return self.model_dump_json(
            by_alias=by_alias, exclude_unset=exclude_unset, **kwargs
        )

    def model_dump(
        self, *, by_alias: bool = True, exclude_unset: bool = True, **kwargs: Any
    ) -> Dict[str, Any]:
        return super().model_dump(
            by_alias=by_alias, exclude_unset=exclude_unset, **kwargs
        )

    def model_dump_json(
        self, *, by_alias: bool = True, exclude_unset: bool = True, **kwargs: Any
    ) -> str:
        return super().model_dump_json(
            by_alias=by_alias, exclude_unset=exclude_unset, **kwargs
        )


class Provider(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#provider-object
    """

    name: str = Field(..., min_length=1)
    description: Optional[str] = None
    roles: Optional[List[str]] = None
    url: Optional[str] = None


class StacCommonMetadata(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/common-metadata.md
    """

    # Basic
    title: Optional[str] = None
    description: Optional[str] = None
    # Date and Time
    datetime: Optional[UtcDatetime] = Field(...)
    created: Optional[UtcDatetime] = None
    updated: Optional[UtcDatetime] = None
    # Date and Time Range
    start_datetime: Optional[UtcDatetime] = None
    end_datetime: Optional[UtcDatetime] = None
    # Licensing
    license: Optional[str] = None
    # Provider
    providers: Optional[List[Provider]] = None
    # Instrument
    platform: Optional[str] = None
    instruments: Optional[List[str]] = None
    constellation: Optional[str] = None
    mission: Optional[str] = None
    gsd: Optional[float] = Field(None, gt=0)

    @model_validator(mode="after")
    def validate_datetime_or_start_end(self) -> Self:
        # When datetime is null, start_datetime and end_datetime must be specified
        if not self.datetime and (not self.start_datetime or not self.end_datetime):
            raise ValueError(
                "start_datetime and end_datetime must be specified when datetime is null"
            )

        return self

    @model_validator(mode="after")
    def validate_start_end(self) -> Self:
        # Using one of start_datetime or end_datetime requires the use of the other
        if (self.start_datetime and not self.end_datetime) or (
            not self.start_datetime and self.end_datetime
        ):
            raise ValueError(
                "use of start_datetime or end_datetime requires the use of the other"
            )
        return self


class Asset(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/item-spec.md#asset-object
    """

    href: str = Field(..., min_length=1)
    type: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    roles: Optional[List[str]] = None

    model_config = ConfigDict(
        populate_by_name=True, use_enum_values=True, extra="allow"
    )


def str_to_datetimes(value: str) -> List[Optional[dt]]:
    # Split on "/" and replace no value or ".." with None
    values = [v if v and v != ".." else None for v in value.split("/")]

    # Cast because pylance gets confused by the type adapter and annotated type
    dates = cast(
        List[Optional[dt]],
        [
            # Use the type adapter to validate the datetime strings, strict is necessary
            # due to pydantic issues #8736 and #8762
            SearchDatetime.validate_strings(v, strict=True) if v else None
            for v in values
        ],
    )
    return dates


def validate_datetime(v: Optional[str]) -> Optional[str]:
    """Validate Datetime value."""
    if v is not None:
        dates = str_to_datetimes(v)

        # If there are more than 2 dates, it's invalid
        if len(dates) > 2:
            raise ValueError(
                "Invalid datetime range. Too many values. Must match format: {begin_date}/{end_date}"
            )

        # If there is only one date, duplicate to use for both start and end dates
        if len(dates) == 1:
            dates = [dates[0], dates[0]]

        # If there is a start and end date, check that the start date is before the end date
        if dates[0] and dates[1] and dates[0] > dates[1]:
            raise ValueError(
                "Invalid datetime range. Begin date after end date. "
                "Must match format: {begin_date}/{end_date}"
            )

    return v


def validate_bbox(v: Optional[BBox]) -> Optional[BBox]:
    """Validate BBOX value."""
    if v:
        # Validate order
        if len(v) == 4:
            xmin, ymin, xmax, ymax = cast(Tuple[int, int, int, int], v)

        elif len(v) == 6:
            xmin, ymin, min_elev, xmax, ymax, max_elev = cast(
                Tuple[int, int, int, int, int, int], v
            )
            if max_elev < min_elev:
                raise ValueError(
                    "Maximum elevation must greater than minimum elevation"
                )
        else:
            raise ValueError("Bounding box must have 4 or 6 coordinates")

        # Validate against WGS84
        if xmin < -180 or ymin < -90 or xmax > 180 or ymax > 90:
            raise ValueError("Bounding box must be within (-180, -90, 180, 90)")

        if xmax < xmin and (xmax > 0 or xmin < 0):
            raise ValueError(
                f"Maximum longitude ({xmax}) must be greater than minimum ({xmin}) longitude when not crossing the Antimeridian"
            )

        if ymax < ymin:
            raise ValueError(
                f"Maximum latitude ({ymax}) must be greater than minimum latitude  ({ymin})"
            )

    return v



from typing import Any, Dict, List, Optional

from geojson_pydantic import Feature
from pydantic import AnyUrl, ConfigDict, Field, model_serializer, model_validator

from stac_pydantic.links import Links
from stac_pydantic.shared import SEMVER_REGEX, Asset, StacBaseModel, StacCommonMetadata
from stac_pydantic.version import STAC_VERSION


class ItemProperties(StacCommonMetadata):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/item-spec.md#properties-object
    """

    model_config = ConfigDict(extra="allow")


class Item(Feature, StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/item-spec.md
    """

    id: str = Field(..., alias="id", min_length=1)
    stac_version: str = Field(STAC_VERSION, pattern=SEMVER_REGEX)
    properties: ItemProperties
    assets: Dict[str, Asset]
    links: Links
    stac_extensions: Optional[List[AnyUrl]] = []
    collection: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def validate_bbox(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if isinstance(values, dict):
            if values.get("geometry") and values.get("bbox") is None:
                raise ValueError("bbox is required if geometry is not null")
        return values

    # https://github.com/developmentseed/geojson-pydantic/issues/147
    @model_serializer(mode="wrap")
    def _serialize(self, handler):
        data = handler(self)
        for field in self.__geojson_exclude_if_none__:
            if field in data and data[field] is None:
                del data[field]
        return data



import json
from typing import Literal

import pytest
from pydantic import ConfigDict, ValidationError
from shapely.geometry import shape

from stac_pydantic import Collection, Item, ItemProperties
from stac_pydantic.collection import SpatialExtent, TimeInterval
from stac_pydantic.extensions import _fetch_and_cache_schema, validate_extensions
from stac_pydantic.links import Link, Links
from stac_pydantic.shared import MimeTypes, StacCommonMetadata

from .conftest import dict_match, request

COLLECTION = "landsat-collection.json"
ITEM_COLLECTION = "itemcollection-sample-full.json"
SINGLE_FILE_STAC = "example-search.json"

# ASSET_EXTENSION = f"https://raw.githubusercontent.com/radiantearth/stac-spec/v{STAC_VERSION}/extensions
# /asset/examples/example-landsat8.json"
# COLLECTION_ASSET_EXTENSION = f"https://raw.githubusercontent.com/radiantearth/stac-spec/v{STAC_VERSION}
# /extensions/collection-assets/examples/example-esm.json"
DATACUBE_EXTENSION = "example-item_datacube-extension.json"
EO_EXTENSION = "example-landsat8_eo-extension.json"
ITEM_ASSET_EXTENSION = "example-landsat8_item-assets-extension.json"
LABEL_EXTENSION = "roads_item.json"
POINTCLOUD_EXTENSION = "example-autzen.json"
PROJ_EXTENSION = "example-landsat8_projection-extension.json"
SAR_EXTENSION = "sentinel1_sar-extension.json"
SAT_EXTENSION = "example-landsat8_sat-extension.json"
SCIENTIFIC_EXTENSION = "example-item_sci-extension.json"
VERSION_EXTENSION_ITEM = "example-item_version-extension.json"
VERSION_EXTENSION_COLLECTION = "example-collection_version-extension.json"
VIEW_EXTENSION = "example-landsat8_view-extension.json"
DATETIME_RANGE = "datetimerange.json"

ITEM_GEOMETRY_NULL = "example-item_geometry-null.json"


@pytest.mark.parametrize(
    "infile",
    [
        EO_EXTENSION,
        POINTCLOUD_EXTENSION,
        SAT_EXTENSION,
        VIEW_EXTENSION,
        SCIENTIFIC_EXTENSION,
        DATACUBE_EXTENSION,
        DATETIME_RANGE,
    ],
)
def test_item_extensions(infile):
    test_item = request(infile)
    valid_item = Item(**test_item).model_dump()
    dict_match(test_item, valid_item)


def test_sar_extensions() -> None:
    test_item = request(SAR_EXTENSION)
    valid_item = Item(**test_item).model_dump()
    dict_match(test_item, valid_item)


def test_proj_extension() -> None:
    # The example item uses an invalid band name
    test_item = request(PROJ_EXTENSION)
    test_item["stac_extensions"][1] = (
        "https://raw.githubusercontent.com/stac-extensions/projection/v1.0.0/json-schema/schema.json"
    )
    test_item["assets"]["B8"]["eo:bands"][0]["common_name"] = "pan"

    valid_item = Item.model_validate(test_item).model_dump()
    dict_match(test_item, valid_item)


def test_version_extension_item() -> None:
    test_item = request(VERSION_EXTENSION_ITEM)
    valid_item = Item(**test_item).model_dump()
    dict_match(test_item, valid_item)


def test_version_extension_collection() -> None:
    test_coll = request(VERSION_EXTENSION_COLLECTION)
    valid_coll = Collection(**test_coll).model_dump(mode="json")
    dict_match(test_coll, valid_coll)


def test_item_assets_extension() -> None:
    test_coll = request(ITEM_ASSET_EXTENSION)
    valid_coll = Collection(**test_coll).model_dump(mode="json")
    dict_match(test_coll, valid_coll)


def test_label_extension() -> None:
    test_item = request(LABEL_EXTENSION)

    # This example contains an invalid geometry (linear ring does not close)
    coords = test_item["geometry"]["coordinates"]
    assert len(coords[0]) == 4
    coords[0].append(coords[0][0])
    test_item["geometry"]["coordinates"] = coords

    valid_item = Item(**test_item).model_dump()
    dict_match(test_item, valid_item)


def test_explicit_extension_validation() -> None:
    test_item = request(EO_EXTENSION)

    # This item implements the eo and view extensions
    assert test_item["stac_extensions"][:-1] == [
        "https://stac-extensions.github.io/eo/v1.0.0/schema.json",
        "https://stac-extensions.github.io/view/v1.0.0/schema.json",
    ]

    validate_extensions(test_item)


def test_extension_validation_schema_cache() -> None:
    # Defines 3 extensions, but one is a non-existing URL
    test_item = request(EO_EXTENSION)

    _fetch_and_cache_schema.cache_clear()

    assert not validate_extensions(test_item)
    assert _fetch_and_cache_schema.cache_info().hits == 0
    assert _fetch_and_cache_schema.cache_info().misses == 3

    assert not validate_extensions(test_item)
    assert _fetch_and_cache_schema.cache_info().hits == 2
    # The non-existing URL will have failed, hence retried
    assert _fetch_and_cache_schema.cache_info().misses == 4


@pytest.mark.parametrize(
    "infile,model",
    [(EO_EXTENSION, Item), (COLLECTION, Collection)],
)
def test_to_json(infile, model):
    test_item = request(infile)
    validated = model(**test_item)
    dict_match(
        json.loads(validated.model_dump_json()), validated.model_dump(mode="json")
    )


def test_item_to_json() -> None:
    test_item = request(EO_EXTENSION)
    item = Item(**test_item)
    dict_match(json.loads(item.model_dump_json()), item.model_dump())


def test_invalid_geometry() -> None:
    test_item = request(EO_EXTENSION)

    # Remove the last coordinate
    test_item["geometry"]["coordinates"][0].pop(-1)

    with pytest.raises(ValidationError):
        Item(**test_item)


def test_asset_extras() -> None:
    test_item = request(EO_EXTENSION)
    for asset in test_item["assets"]:
        test_item["assets"][asset]["foo"] = "bar"

    item = Item(**test_item)
    for _, asset in item.assets.items():
        assert asset.foo == "bar"


def test_geo_interface() -> None:
    test_item = request(EO_EXTENSION)
    item = Item(**test_item)
    geom = shape(item.geometry)
    test_item["geometry"] = geom
    Item(**test_item)


@pytest.mark.parametrize(
    "args",
    [
        {"datetime": "2024-01-01T00:00:00Z"},
        {
            "datetime": None,
            "start_datetime": "2024-01-01T00:00:00Z",
            "end_datetime": "2024-01-02T00:00:00Z",
        },
        {
            "datetime": "2024-01-01T00:00:00Z",
            "start_datetime": "2024-01-01T00:00:00Z",
            "end_datetime": "2024-01-02T00:00:00Z",
        },
    ],
)
def test_stac_common_dates(args) -> None:
    StacCommonMetadata(**args)


def test_stac_null_datetime_required() -> None:
    with pytest.raises(ValidationError):
        StacCommonMetadata(
            **{
                "start_datetime": "2024-01-01T00:00:00Z",
                "end_datetime": "2024-01-02T00:00:00Z",
            }
        )


@pytest.mark.parametrize(
    "args",
    [
        {"datetime": None},
        {"datetime": None, "start_datetime": "2024-01-01T00:00:00Z"},
        {"datetime": None, "end_datetime": "2024-01-01T00:00:00Z"},
    ],
)
def test_stac_common_no_dates(args) -> None:
    with pytest.raises(
        ValueError,
        match="start_datetime and end_datetime must be specified when datetime is null",
    ):
        StacCommonMetadata(**args)


@pytest.mark.parametrize(
    "args",
    [
        {"datetime": "2024-01-01T00:00:00Z", "start_datetime": "2024-01-01T00:00:00Z"},
        {"datetime": "2024-01-01T00:00:00Z", "end_datetime": "2024-01-01T00:00:00Z"},
    ],
)
def test_stac_common_start_and_end(args) -> None:
    with pytest.raises(
        ValueError,
        match="use of start_datetime or end_datetime requires the use of the other",
    ):
        StacCommonMetadata(**args)


def test_declared_model() -> None:
    class TestProperties(ItemProperties):
        foo: str
        bar: int

        model_config = ConfigDict(
            populate_by_name=True,
            alias_generator=lambda field_name: f"test:{field_name}",
        )

    class TestItem(Item):
        properties: TestProperties

    test_item = request(EO_EXTENSION)
    del test_item["stac_extensions"]
    test_item["properties"] = {
        "datetime": test_item["properties"]["datetime"],
        "test:foo": "mocked",
        "test:bar": 1,
    }

    valid_item = TestItem(**test_item).model_dump()

    assert "test:foo" in valid_item["properties"]
    assert "test:bar" in valid_item["properties"]


def test_item_factory_custom_base() -> None:
    class TestProperties(ItemProperties):
        foo: Literal["bar"] = "bar"

    class TestItem(Item):
        properties: TestProperties

    test_item = request(EO_EXTENSION)

    model = TestItem(**test_item)
    assert model.properties.foo == "bar"


def test_serialize_namespace() -> None:
    test_item = request(SAR_EXTENSION)
    valid_item = Item(**test_item)
    assert "sar:instrument_mode" in valid_item.model_dump()["properties"]


def test_excludes() -> None:
    test_item = request(EO_EXTENSION)
    valid_item = Item(**test_item).model_dump(
        by_alias=True, exclude_unset=True, exclude={"properties": {"bands"}}
    )
    assert "eo:bands" not in valid_item["properties"]


def test_validate_extensions() -> None:
    test_item = request(SAR_EXTENSION)
    assert validate_extensions(test_item)


def test_validate_extensions_reraise_exception() -> None:
    test_item = request(EO_EXTENSION)
    del test_item["properties"]["datetime"]

    with pytest.raises(ValidationError):
        Item.model_validate(test_item)
        validate_extensions(test_item, reraise_exception=True)


def test_validate_extensions_rfc3339_with_partial_seconds() -> None:
    test_item = request(SAR_EXTENSION)
    test_item["properties"]["updated"] = "2018-10-01T01:08:32.033Z"
    assert validate_extensions(test_item)


@pytest.mark.parametrize("url,cls", [[EO_EXTENSION, Item], [COLLECTION, Collection]])
def test_extension(url, cls):
    test_data = request(url)
    test_data["stac_extensions"].append("https://foo")
    model = cls.model_validate(test_data)
    assert "https://foo/" in list(map(str, model.stac_extensions))


def test_resolve_link() -> None:
    link = Link(href="/hello/world", type=MimeTypes.jpeg, rel="test")
    link.resolve(base_url="http://base_url.com")
    assert link.href == "http://base_url.com/hello/world"


def test_resolve_links() -> None:
    links = Links.model_validate(
        [Link(href="/hello/world", type=MimeTypes.jpeg, rel="test")]
    )
    links.resolve(base_url="http://base_url.com")
    for link in links.link_iterator():
        assert link.href == "http://base_url.com/hello/world"


def test_geometry_null_item() -> None:
    test_item = request(ITEM_GEOMETRY_NULL)
    valid_item = Item(**test_item).model_dump()
    dict_match(test_item, valid_item)


def test_item_bbox_validation() -> None:
    test_item = request(LABEL_EXTENSION)
    test_item["bbox"] = None
    with pytest.raises(ValueError, match="bbox is required if geometry is not null"):
        Item(**test_item)


@pytest.mark.parametrize(
    "interval",
    [
        # interval should at least have 1 value
        [],
        # interval item should be an array of 2 datetime
        ["2023-01-01T00:00:00Z", "2024-01-01T00:00:00Z"],
        # interval item should not be null
        [[]],
        # interval item should have 2 values
        [["2023-01-01T00:00:00Z", "2023-01-02T00:00:00Z", "2024-01-01T00:00:00Z"]],
        # invalid datetime values
        [[None, "yo"]],
        [["yo", None]],
        [["yo", "yo"]],
        # wrong datetime order
        [["2024-01-01T00:00:00Z", "2023-01-01T00:00:00Z"]],
        # sub-sequent starttime before overall starttime
        [
            ["2023-01-01T00:00:00Z", "2024-01-01T00:00:00Z"],
            ["2022-01-31T00:00:00Z", "2024-01-01T00:00:00Z"],
        ],
        # sub-sequent endtime after overall endtime
        [
            ["2023-01-01T00:00:00Z", "2024-01-01T00:00:00Z"],
            ["2023-01-31T00:00:00Z", "2024-01-02T00:00:00Z"],
        ],
        # sub-sequent starttime after sub-sequent endtime
        [
            ["2023-01-01T00:00:00Z", "2024-01-01T00:00:00Z"],
            ["2023-01-31T00:00:00Z", "2023-01-01T00:00:00Z"],
        ],
    ],
)
def test_time_intervals_invalid(interval) -> None:
    """Check Time Interval model."""
    with pytest.raises(ValidationError):
        TimeInterval(interval=interval)


@pytest.mark.parametrize(
    "interval",
    [
        # Open date range to both sides is supported but not recommended.
        [[None, None]],
        [["2024-01-01T00:00:00Z", None]],
        [[None, "2024-01-01T00:00:00Z"]],
        [["2023-01-01T00:00:00Z", "2024-01-01T00:00:00Z"]],
        [
            ["2023-01-01T00:00:00Z", "2024-01-01T00:00:00Z"],
            ["2023-01-31T00:00:00Z", "2024-01-01T00:00:00Z"],
        ],
    ],
)
def test_time_intervals_valid(interval) -> None:
    """Check Time Interval model."""
    assert TimeInterval(interval=interval)


@pytest.mark.parametrize(
    "bboxes",
    [
        # invalid Y order
        [[0, 1, 1, 0]],
        # invalid X order (if crossing Antimeridian limit, xmin > 0)
        [[-169, 0, -170, 1]],
        # invalid X order (if crossing Antimeridian limit, xmax < 0)
        [[170, 0, 169, 1]],
        # sub-sequent crossing Y
        [[0, 0, 2, 2], [0.5, 0.5, 2.0, 2.5]],
        # sub-sequent crossing X
        [[0, 0, 2, 2], [0.5, 0.5, 2.5, 2.0]],
        # sub-sequent crossing Antimeridian limit
        [[0, 0, 2, 2], [1, 0, -179, 1]],
        # both crossing Antimeridian limit but sub-sequent cross has min lat -176 > -178
        [[2, 0, -178, 2], [1, 0, -176, 1]],
        # sub-sequent cross Antimeridian but not the overall
        [[0, 0, 2, 2], [1, 0, -176, 1]],
        # overall crossing and sub-sequent not within bounds
        [[2, 0, -178, 2], [-179, 0, -176, 1]],
        # overall crossing and sub-sequent not within bounds
        [[2, 0, -178, 2], [1, 0, 3, 1]],
    ],
)
def test_spatial_intervals_invalid(bboxes) -> None:
    """Check invalid Spatial Interval model."""
    with pytest.raises(ValidationError):
        SpatialExtent(bbox=bboxes)


@pytest.mark.parametrize(
    "bboxes",
    [
        [[0, 0, 1, 1]],
        # Same on both side
        [[0, 0, 2, 2], [0, 0, 2, 2]],
        [[0, 0, 2, 2], [0.5, 0.5, 1.5, 1.5]],
        # crossing Antimeridian limit
        [[2, 0, -178, 2]],
        # Case 1: overall crossing Antimeridian, sub-sequent bbox not crossing (but within overall right part)
        [[2, 0, -178, 2], [-179, 0, -178, 1]],
        # Case 2: overall crossing Antimeridian, sub-sequent bbox not crossing (but within overall left part)
        [[2, 0, -178, 2], [179, 0, 180, 1]],
        # Case 3: overall and sub-sequent crossing Antimeridian
        [[2, 0, -178, 2], [3, 0, -179, 1]],
    ],
)
def test_spatial_intervals_valid(bboxes) -> None:
    """Check Spatial Interval model."""
    assert SpatialExtent(bbox=bboxes)
